#include "stdafx.h"

#include <intrin.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>

// Pull in the XMVector math functions
#include <DirectXMath.h>
using namespace DirectX;



// The original version of _mm_ceil_ps2 from http://dss.stephanierct.com/DevBlog/?p=8, now fixed.
inline __m128 old_mm_ceil_ps2(const __m128& a){
	__m128 v0 = _mm_setzero_ps();
	__m128 v1 = _mm_cmpeq_ps(v0, v0);
	__m128 vd5 = *(__m128*)&_mm_slli_epi32(*(__m128i*)&v1, 26);
	vd5 = *(__m128*)&_mm_srli_epi32(*(__m128i*)&vd5, 2); // make a 0.5f value
	__m128 vMinusD5 = _mm_add_ps(a, vd5);
	__m128i i = _mm_cvtps_epi32(vMinusD5);
	__m128 r = _mm_cvtepi32_ps(i);
	return r;
}

// The fixed version of _mm_ceil_ps2 from http://dss.stephanierct.com/DevBlog/?p=8.
inline __m128 new_mm_ceil_ps2(const __m128& x){
	__m128i v0 = _mm_setzero_si128();
	__m128i v1 = _mm_cmpeq_epi32(v0, v0);
	__m128i ji = _mm_srli_epi32(v1, 25);
	__m128 j = *(__m128*)&_mm_slli_epi32(ji, 23); //create vector 1.0f
	__m128i i = _mm_cvttps_epi32(x);
	__m128 fi = _mm_cvtepi32_ps(i);
	__m128 igx = _mm_cmplt_ps(fi, x);
	j = _mm_and_ps(igx, j);
	return _mm_add_ps(fi, j);
}

// Wrapper class to handle large numbers from http://dss.stephanierct.com/DevBlog/?p=8 (doesn't handle NaNs).
template< __m128 (FuncT)(const __m128&) >
inline __m128 _mm_safeInt_ps(const __m128& a){
	__m128 v8388608 = *(__m128*)&_mm_set1_epi32(0x4b000000);                        //vector with value 8388608
	__m128 aAbs = _mm_and_ps(a, *(__m128*)&_mm_set1_epi32(0x7fffffff));             //Abs(a)
	__m128 aMask = _mm_cmpgt_ps(aAbs, v8388608);                                    //if Abs(a) > 8388608
	// select a if greater then 8388608.0f, otherwise select the result of FuncT
	__m128 r = _mm_xor_ps(_mm_and_ps(aMask, a), _mm_andnot_ps(aMask, FuncT(a)));
	return r;
}

float test_new_full_mm_ceil_ps2(float f)
{
	__m128 input = { f, 0, 0, 0 };
	__m128 result = _mm_safeInt_ps<new_mm_ceil_ps2>(input);
	return result.m128_f32[0];
}

// Wrapper class to handle large numbers from http://dss.stephanierct.com/DevBlog/?p=8 (with my fix to handle NaNs).
template< __m128 (FuncT)(const __m128&) >
inline __m128 _mm_fixed_safeInt_ps(const __m128& a){
	__m128 v8388608 = *(__m128*)&_mm_set1_epi32(0x4b000000);
	__m128 aAbs = _mm_and_ps(a, *(__m128*)&_mm_set1_epi32(0x7fffffff));
	// In order to handle NaNs correctly we need to use le instead of gt.
	// Using le ensures that the bitmask is clear for large numbers *and*
	// NaNs, whereas gt ensures that the bitmask is set for large numbers
	// but not for NaNs.
	__m128 aMask = _mm_cmple_ps(aAbs, v8388608);   //if Abs(a) <= 8388608
	// select a if greater then 8388608.0f, otherwise select the result of
	// FuncT. Note that 'and' and 'andnot' were reversed because the
	// meaning of the bitmask has been reversed.
	__m128 r = _mm_xor_ps(_mm_andnot_ps(aMask, a), _mm_and_ps(aMask, FuncT(a)));
	return r;
}

float test_new_fixed_full_mm_ceil_ps2(float f)
{
	__m128 input = { f, 0, 0, 0 };
	__m128 result = _mm_fixed_safeInt_ps<new_mm_ceil_ps2>(input);
	return result.m128_f32[0];
}

float test_new_mm_ceil_ps2(float f)
{
	__m128 input = { f, 0, 0, 0 };
	__m128 result = new_mm_ceil_ps2(input);
	return result.m128_f32[0];
}

float old_mm_ceil_ps2(float f)
{
	__m128 input = { f, 0, 0, 0 };
	__m128 result = old_mm_ceil_ps2(input);
	return result.m128_f32[0];
}



// Extracted from the 'ceil' function in:
// https://github.com/LiraNuna/glsl-sse2/blob/master/source/vec4.h
inline const __m128 lira_ceil(const __m128 &v) {
	__m128 m = _mm_cmpunord_ps(v,
		_mm_cmpge_ps(_mm_andnot_ps(_mm_set1_ps(-0.0f), v),
		_mm_set1_ps(8388608.0f)));
	return _mm_or_ps(_mm_andnot_ps(m, _mm_cvtepi32_ps(_mm_cvtps_epi32(
		_mm_add_ps(v, _mm_set1_ps(0.5f))))),
		_mm_and_ps(m, v));
}

float test_lira_ceil(float f)
{
	__m128 input = { f, 0, 0, 0 };
	__m128 result = lira_ceil(input);
	return result.m128_f32[0];
}



// The XMVector library used is chosen by whether you compile for VS 2013 or earlier.
// In Project Properties, General, set Platform Toolset to either VS 2013 or some
// earlier compiler in order to test version 3.06 (mostly works) versus 3.03 (mostly
// doesn't work) of the math library.

#define STRINGIZEX(x) #x
#define STRINGIZE(x) _STRINGIZEX(x)

float XMVectorCeiling_Wrapper(float f)
{
	FXMVECTOR input = { f, 0, 0, 0 };
	FXMVECTOR result = XMVectorCeiling(input);
	return result.m128_f32[0];
}

float XMVectorFloor_Wrapper(float f)
{
	FXMVECTOR input = { f, 0, 0, 0 };
	FXMVECTOR result = XMVectorFloor(input);
	return result.m128_f32[0];
}

float XMVectorRound_Wrapper(float f)
{
	FXMVECTOR input = { f, 0, 0, 0 };
	FXMVECTOR result = XMVectorRound(input);
	return result.m128_f32[0];
}



int main()
{
	TimeIt timer;
	// This is the biggest number that can be represented in
	// both float and int32_t. It's 2^31-128.
	Float_t maxfloatasint(2147483520.0f);
	Float_t one(1.0f);
	const uint32_t signBit = 0x80000000;

	//ExhaustiveTest(0, (uint32_t)maxfloatasint.i, old_mm_ceil_ps2, ceil, "old _mm_ceil_ps2");
	//ExhaustiveTest(signBit, signBit | maxfloatasint.i, old_mm_ceil_ps2, ceil, "old _mm_ceil_ps2");

	//ExhaustiveTest(0, 0xFFFFFFFF, XMVectorCeiling_Wrapper, ceil, "XMVectorCeiling" STRINGIZE(DIRECTX_MATH_VERSION));
	//ExhaustiveTest(0, 0xFFFFFFFF, XMVectorFloor_Wrapper, floor, "XMVectorFloor" STRINGIZE(DIRECTX_MATH_VERSION));
	//ExhaustiveTest(0, 0xFFFFFFFF, XMVectorRound_Wrapper, round, "XMVectorRound" STRINGIZE(DIRECTX_MATH_VERSION));

	//ExhaustiveTest(0, 0xFFFFFFFF, test_new_full_mm_ceil_ps2, ceil, "new full mm_ceil_ps2");
	//ExhaustiveTest(0, 0xFFFFFFFF, test_new_fixed_full_mm_ceil_ps2, ceil, "new fixed full mm_ceil_ps2");

	//ExhaustiveTest(0, 0xFFFFFFFF, test_lira_ceil, ceil, "lira_ceil");
}
