#include "stdafx.h"

// Define this to get ExhaustiveTest to check whether the sign of zero is correct in the results.
#define CHECK_SIGN_OF_ZERO

const int kMaxDisplayedErrors = 10;
const int kMaxDisplayedCorrect = 4;

// Pass in a uint32_t range of float representations to test.
// start and stop are inclusive. Pass in 0, 0xFFFFFFFF to scan all
// floats. The floats are iterated through by incrementing
// their integer representation.
void ExhaustiveTest(uint32_t start, uint32_t stop, Transform TestFunc,
	Transform RefFunc, const char* desc)
{
	Float_t start_f, stop_f;
	start_f.i = (int32_t)start;
	stop_f.i = (int32_t)stop;
	printf("Testing %s from %.9g (%u) to %.9g (%u) (inclusive).\n", desc, start_f.f, start, stop_f.f, stop);
	int count = 0;
	long long nErrors = 0;
	// Use long long to let us loop over all positive integers.
	long long i = start;
	long long totalTested = (long long)stop - start + 1;
	printf("%lld numbers will be tested.\n", totalTested);
	while (i <= stop)
	{
		Float_t input;
		input.i = (int32_t)i;
		Float_t testValue = TestFunc(input.f);
		Float_t refValue = RefFunc(input.f);
		// If the results don't match then report an error.
#ifdef CHECK_SIGN_OF_ZERO
		// If we check the internal representations then we will detect when
		// +0 and -0 are mismatched. This an additional billion failures
		// for many of these functions, and is too fussy for many uses.
		if (testValue.i != refValue.i &&
#else
		// Comparing the float values means that +0 and -0 will be treated
		// as equal.
		if (testValue.f != refValue.f &&
#endif
			// If both results are NaNs then we treat that as a match.
			(testValue.f == testValue.f || refValue.f == refValue.f))
		{
			++nErrors;
			if (nErrors <= kMaxDisplayedErrors)
				printf("Input %.9g, expected %.9g, got %1.9g        \n",
				input.f, refValue.f, testValue.f);
		}

		++i;

		// Occasionally print out a progress message
		if (++count > 1000000)
		{
			count = 0;
			// Use %1.8e here so it consistently overwrites itself.
			printf("At %1.8e, %lld errors\r", input.f, nErrors);
		}
	}

	printf("%lld errors (%1.1f%%) detected in %s.\n\n", nErrors, nErrors * 100.0 / totalTested, desc);
}

typedef float(*Transform)(float);

// Pass in a uint32_t range of float representations to test.
// start and stop are inclusive. Pass in 0, 0xFFFFFFFF to scan all
// floats. The floats are iterated through by incrementing
// their integer representation.
void ExhaustiveTestSimple(uint32_t start, uint32_t stop, Transform TestFunc,
	Transform RefFunc, const char* desc)
{
	printf("Testing %s from %u to %u (inclusive).\n", desc, start, stop);
	// Use long long to let us loop over all positive integers.
	long long i = start;
	while (i <= stop)
	{
		Float_t input;
		input.i = (int32_t)i;
		Float_t testValue = TestFunc(input.f);
		Float_t refValue = RefFunc(input.f);
		// If the results don't match then report an error.
		if (testValue.f != refValue.f &&
			// If both results are NaNs then we treat that as a match.
			(testValue.f == testValue.f || refValue.f == refValue.f))
		{
			printf("Input %.9g, expected %.9g, got %1.9g        \n",
				input.f, refValue.f, testValue.f);
		}

		++i;
	}
}
